"""Init Optimize Submodule."""

from ..core import anneal_enumerations
from .bonsai import bonsai_algorithm
from .enumeration.cost_functions import (
    distance_squared,
    minimise_mi_distance,
    pauli_weighted_norm,
)
from .enumeration.evolutionary import lambda_plus_mu
from .huffman import huffman_ternary_tree
from .rett import reduced_entanglement_ternary_tree

__all__ = [
    "lambda_plus_mu",
    "minimise_mi_distance",
    "distance_squared",
    "pauli_weighted_norm",
    "anneal_enumerations",
    "bonsai_algorithm",
    "huffman_ternary_tree",
    "reduced_entanglement_ternary_tree",
]



import numpy as np
import pytest
import scipy as sp
from ferrmion.encode.ternary_tree import TernaryTree, TTNode, JW, JordanWigner, BK, BravyiKitaev, JKMN, ParityEncoding
from ferrmion.utils import symplectic_hash, symplectic_unhash
from openfermion import QubitOperator, get_sparse_operator
from openfermion.ops import InteractionOperator
from openfermion.transforms import jordan_wigner
from ferrmion.hamiltonians import molecular_hamiltonian


@pytest.fixture
def six_mode_tree():
    return TernaryTree(n_modes=6, root_node=TTNode())

def test_standard_encoding_functions(six_mode_tree):
    # Test function aliases
    assert JW(6) == JordanWigner(6)
    assert BK(6) == BravyiKitaev(6)

    # Test TT function aliases
    assert six_mode_tree.JW() == JW(6)
    assert six_mode_tree.JordanWigner() == JordanWigner(6)
    assert six_mode_tree.BK() == BK(6)
    assert six_mode_tree.BravyiKitaev() == BravyiKitaev(6)
    assert six_mode_tree.JKMN() == JKMN(6)
    assert six_mode_tree.ParityEncoding() == ParityEncoding(6)

    # Test inequality by type
    assert JW(6) != BK(6)
    assert JW(6) != JKMN(6)
    assert JW(6) != ParityEncoding(6)
    assert BK(6) != JKMN(6)
    assert BK(6) != ParityEncoding(6)
    assert JKMN(6) != ParityEncoding(6)

    # Test inequality
    assert JW(6) != JW(5)
    assert JW(6) != JW
    assert JW(6) != "JW(6)"

    jw_different_enumeration = JW(6)
    jw_different_enumeration.enumeration_scheme["z"] = JW(6).enumeration_scheme["zz"]
    jw_different_enumeration.enumeration_scheme["zz"] = JW(6).enumeration_scheme["z"]
    assert JW(6) != jw_different_enumeration

def test_ttnode():
    root = TTNode()
    child = root.add_child("x", root_path=0)
    child = child.add_child("y", root_path=1)
    child = child.add_child("z", root_path=2)
    assert child.parent.parent.parent == root
    assert child.parent.parent == root.x
    assert child.parent == root.x.y
    assert root.root_path is None
    assert root.x.root_path == 0
    assert root.as_dict() == {
        "x": {
            "x": {},
            "y": {"x": {}, "y": {}, "z": {"x": {}, "y": {}, "z": {}}},
            "z": {},
        },
        "y": {},
        "z": {},
    }
    assert root.branch_strings == {
        "xx",
        "xyx",
        "xyy",
        "xyzx",
        "xyzy",
        "xyzz",
        "xz",
        "y",
        "z",
    }

def test_enumeration_scheme(six_mode_tree):
    assert six_mode_tree.default_enumeration_scheme() == {'':(0,0)}
    jkmn = six_mode_tree.JKMN()
    assert jkmn.default_enumeration_scheme() == {'': (0, 0), 'x': (1, 1), 'y': (2, 2), 'z': (3, 3), 'xx': (4, 4), 'xy': (5, 5)}

    # Not enough qubit labels
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {'': (0, 0), 'x': (1, 1), 'y': (2, 2), 'z': (3, 3), 'xx': (4, 4), 'xy': (5, 4)}
    assert "Invalid qubit labels" in str(exc.value)

    # Not enough mode labels
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {'': (0, 0), 'x': (1, 1), 'y': (2, 2), 'z': (3, 3), 'xx': (5, 4), 'xy': (5, 5)}
    assert "Invalid mode labels" in str(exc.value)

    # Qubit label not in range
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {'': (0, 6), 'x': (1, 1), 'y': (2, 2), 'z': (3, 3), 'xx': (4, 4), 'xy': (5, 5)}
    assert "Invalid qubit labels" in str(exc.value)

    # Mode label not in range
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {'': (6, 0), 'x': (1, 1), 'y': (2, 2), 'z': (3, 3), 'xx': (4, 4), 'xy': (5, 5)}
    assert "Invalid mode labels" in str(exc.value)

    jkmn.enumeration_scheme = {'': (3, 1), 'x': (2, 5), 'y': (0, 3), 'z': (1, 4), 'xx': (4, 2), 'xy': (5, 0)}
    assert np.all(jkmn._build_symplectic_matrix()[1] == np.array([[False,  True, False,  True, False, False, False,  True, False,
         True, False, False],
       [False,  True, False,  True, False, False, False,  True, False,
        False, False, False],
       [False, False, False, False,  True, False, False,  True, False,
        False, False, False],
       [False, False, False, False,  True, False, False,  True, False,
        False,  True, False],
       [False,  True, False, False, False,  True, False, False,  True,
        False, False, False],
       [False,  True, False, False, False,  True,  True, False, False,
        False, False,  True],
       [False,  True, False, False, False, False, False, False, False,
        False, False,  True],
       [False,  True, False, False, False, False, False,  True, False,
         True, False, False],
       [False,  True,  True, False, False,  True, False, False, False,
        False, False, False],
       [False,  True,  True, False, False,  True, False, False,  True,
        False, False, False],
       [ True,  True, False, False, False,  True,  True, False, False,
        False, False,  True],
       [ True,  True, False, False, False,  True, False, False, False,
        False, False,  True]]))
def test_ttnode_to_rustworkx(six_mode_tree):
    graph = six_mode_tree.JKMN().root_node.to_rustworkx()
    assert graph.nodes() == ['', 'x', 'y', 'z', 'xx', 'xy']
    assert np.all([*graph.edge_list()] == [(0, 1), (0, 2), (0, 3), (1, 4), (1, 5)])

# def test_jordan_wigner(six_mode_tree):
#     jw = six_mode_tree.JW()
#     assert jw.root_node.branch_strings == {
#         "x",
#         "y",
#         "zx",
#         "zy",
#         "zzx",
#         "zzy",
#         "zzzx",
#         "zzzy",
#         "zzzz",
#     }

#     assert jw.root_node.child_strings == ['', 'z', 'zz', 'zzz']
#     assert len(jw.qubits) == len(jw.root_node.child_strings)

#     assert jw.string_pairs == {'': ('x', 'y'),
#         'z': ('zx', 'zy'),
#         'zz': ('zzx', 'zzy'),
#         'zzz': ('zzzx', 'zzzy')
#     }
#     assert jw.branch_operator_map == {
#         "zx": "ZXII",
#         "zzx": "ZZXI",
#         "zzy": "ZZYI",
#         "zzzx": "ZZZX",
#         "zy": "ZYII",
#         "y": "YIII",
#         "x": "XIII",
#         "zzzy": "ZZZY",
#     }
#     assert jw._build_symplectic_matrix().shape == (2 * len(jw.qubits), 2 * len(jw.qubits))
#     assert np.all(
#         jw._build_symplectic_matrix()
#         == np.array(
#             [
#                 [1, 0, 0, 0, 0, 0, 0, 0],
#                 [1, 0, 0, 0, 1, 0, 0, 0],
#                 [0, 1, 0, 0, 1, 0, 0, 0],
#                 [0, 1, 0, 0, 1, 1, 0, 0],
#                 [0, 0, 1, 0, 1, 1, 0, 0],
#                 [0, 0, 1, 0, 1, 1, 1, 0],
#                 [0, 0, 0, 1, 1, 1, 1, 0],
#                 [0, 0, 0, 1, 1, 1, 1, 1],
#             ]
#         )
#     )
#     assert jw.as_dict() == {'x': {},
#         'y': {},
#         'z': {'x': {},
#         'y': {},
#         'z': {'x': {}, 'y': {}, 'z': {
#             'x': {}, 'y': {}, 'z': {}}
#             }}
#     }
#     assert jw.default_enumeration_scheme() == {'': {'mode': 0, 'qubit': 0},
#         'z': {'mode': 1, 'qubit': 1},
#         'zz': {'mode': 2, 'qubit': 2},
#         'zzz': {'mode': 3, 'qubit': 3}
#     }
#     for line in jw._build_symplectic_matrix():
#         assert np.all(line == symplectic_unhash(symplectic_hash(line), len(line)))


# def test_parity(six_mode_tree):
#     assert six_mode_tree.ParityEncoding().as_dict() == {
#         "x": {
#             "x": {"x": {"x": {}, "y": {}, "z": {}}, "y": {}, "z": {}},
#             "y": {},
#             "z": {},
#         },
#         "y": {},
#         "z": {},
#     }
#     assert six_mode_tree.ParityEncoding().pauli_strings == {
#         "xxxy": "XXXY",
#         "xy": "XYII",
#         "y": "YIII",
#         "xz": "XZII",
#         "xxxx": "XXXX",
#         "xxxz": "XXXZ",
#         "xxz": "XXZI",
#         "xxy": "XXYI",
#     }

#     assert six_mode_tree.ParityEncoding().root_node.branch_strings == {
#         "xxxy",
#         "xy",
#         "y",
#         "xz",
#         "xxxx",
#         "xxxz",
#         "xxz",
#         "z",
#         "xxy",
#     }


def test_bravyi_kitaev(six_mode_tree):
    tt = six_mode_tree.BK()
    assert tt.root_node.branch_strings == {
        "xxzy",
        "xxzx",
        "xxzz",
        "xzx",
        "xzy",
        "xzz",
        "xxy",
        "xxxx",
        "y",
        "xy",
        "z",
        "xxxz",
        "xxxy",
    }

    assert tt.root_node.child_strings == ["", "x", "xx", "xz", "xxx", "xxz"]

    assert tt.as_dict() == {
        "x": {
            "x": {
                "x": {"x": {}, "y": {}, "z": {}},
                "y": {},
                "z": {"x": {}, "y": {}, "z": {}},
            },
            "y": {},
            "z": {"x": {}, "y": {}, "z": {}},
        },
        "y": {},
        "z": {},
    }

    assert tt.default_enumeration_scheme() == {
        "": (0,0),
        "x": (1,1),
        "xx": (2,2),
        "xz": (3,3),
        "xxx": (4,4),
        "xxz": (5,5),
    }

    assert tt.string_pairs == {
        "": ("xzz", "y"),
        "x": ("xxzz", "xy"),
        "xx": ("xxxz", "xxy"),
        "xz": ("xzx", "xzy"),
        "xxx": ("xxxx", "xxxy"),
        "xxz": ("xxzx", "xxzy"),
    }

    assert tt.branch_operator_map == {
        "xxzy": "XXZIIY",
        "xxzx": "XXZIIX",
        "xxzz": "XXZIIZ",
        "xzx": "XZIXII",
        "xzy": "XZIYII",
        "xzz": "XZIZII",
        "xxy": "XXYIII",
        "xxxx": "XXXIXI",
        "y": "YIIIII",
        "xy": "XYIIII",
        "xxxz": "XXXIZI",
        "xxxy": "XXXIYI",
        "z": "ZIIIII",
    }

    assert tt.n_qubits == len(tt.root_node.child_strings)
    assert np.all(
        tt._build_symplectic_matrix()[1]
        == np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0],
                [1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1],
                [1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0],
                [1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0],
                [1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1],
            ],
            dtype=np.int8,
        )
    )

    for line in tt._build_symplectic_matrix()[1]:
        assert np.all(line == symplectic_unhash(symplectic_hash(line), len(line)))


# def test_JKMN(six_mode_tree):
#     assert six_mode_tree.JKMN().as_dict() == {
#         "x": {"x": {}, "y": {}, "z": {}},
#         "y": {"x": {}, "y": {}, "z": {}},
#         "z": {"x": {}, "y": {}, "z": {}},
#     }
#     assert six_mode_tree.JKMN().root_node.branch_strings == {
#         "yz",
#         "yy",
#         "xz",
#         "yx",
#         "xy",
#         "zy",
#         "zx",
#         "zz",
#         "xx",
#     }

#     assert six_mode_tree.JKMN().pauli_strings == {
#         "yz": "YIZI",
#         "yy": "YIYI",
#         "xz": "XZII",
#         "yx": "YIXI",
#         "xy": "XYII",
#         "zy": "ZIIY",
#         "zx": "ZIIX",
#         "xx": "XXII",
#     }

#     assert six_mode_tree.JKMN().string_pairs == {
#         "xz": "yz",
#         "yz": "xz",
#         "xx": "xy",
#         "xy": "xx",
#         "zx": "zy",
#         "zy": "zx",
#         "yx": "yy",
#         "yy": "yx",
#     }


def tests_bonsai_paper_tree():
    tt = TernaryTree(n_modes=11)
    tt = tt.add_node("x")
    tt = tt.add_node("y")
    tt = tt.add_node("z")
    tt = tt.add_node("xx")
    tt = tt.add_node("xy")
    tt = tt.add_node("yx")
    tt = tt.add_node("yy")
    tt = tt.add_node("yz")
    tt = tt.add_node("zz")
    tt = tt.add_node("yzz")
    tt.enumeration_scheme = tt.default_enumeration_scheme()

    assert tt.root_node.branch_strings == {
        "xyz",
        "zzy",
        "yyx",
        "yxz",
        "yzx",
        "yyy",
        "yzzx",
        "xyx",
        "xxx",
        "xxz",
        "yxx",
        "yzy",
        "xyy",
        "xxy",
        "yzzz",
        "yyz",
        "yxy",
        "zx",
        "zzz",
        "xz",
        "yzzy",
        "zzx",
        "zy",
    }

    assert tt.root_node.child_strings == [
        "",
        "x",
        "y",
        "z",
        "xx",
        "xy",
        "yx",
        "yy",
        "yz",
        "zz",
        "yzz",
    ]

    assert tt.as_dict() == {
        "x": {
            "x": {"x": {}, "y": {}, "z": {}},
            "y": {"x": {}, "y": {}, "z": {}},
            "z": {},
        },
        "y": {
            "x": {"x": {}, "y": {}, "z": {}},
            "y": {"x": {}, "y": {}, "z": {}},
            "z": {"x": {}, "y": {}, "z": {"x": {}, "y": {}, "z": {}}},
        },
        "z": {"x": {}, "y": {}, "z": {"x": {}, "y": {}, "z": {}}},
    }

    assert tt.default_enumeration_scheme() == {
        "": (0,0),
        "x": (1,1),
        "y": (2,2),
        "z": (3,3),
        "xx": (4,4),
        "xy": (5,5),
        "yx": (6,6),
        "yy": (7,7),
        "yz": (8,8),
        "zz": (9,9),
        "yzz": (10,10),
    }

    assert tt.string_pairs == {
        "": ("xz", "yzzz"),
        "x": ("xxz", "xyz"),
        "y": ("yyz", "yxz"),
        "z": ("zx", "zy"),
        "xx": ("xxx", "xxy"),
        "xy": ("xyy", "xyx"),
        "yx": ("yxy", "yxx"),
        "yy": ("yyx", "yyy"),
        "yz": ("yzy", "yzx"),
        "zz": ("zzx", "zzy"),
        "yzz": ("yzzy", "yzzx"),
    }

    assert tt.branch_operator_map == {
        "xyz": "XYIIIZIIIII",
        "zzy": "ZIIZIIIIIYI",
        "yyx": "YIYIIIIXIII",
        "yxz": "YIXIIIZIIII",
        "yzx": "YIZIIIIIXII",
        "yyy": "YIYIIIIYIII",
        "yzzx": "YIZIIIIIZIX",
        "xyx": "XYIIIXIIIII",
        "xxx": "XXIIXIIIIII",
        "xxz": "XXIIZIIIIII",
        "yxx": "YIXIIIXIIII",
        "yzy": "YIZIIIIIYII",
        "xyy": "XYIIIYIIIII",
        "xxy": "XXIIYIIIIII",
        "yzzz": "YIZIIIIIZIZ",
        "yyz": "YIYIIIIZIII",
        "yxy": "YIXIIIYIIII",
        "zx": "ZIIXIIIIIII",
        "xz": "XZIIIIIIIII",
        "yzzy": "YIZIIIIIZIY",
        "zzx": "ZIIZIIIIIXI",
        "zy": "ZIIYIIIIIII",
        "zzz": "ZIIZIIIIIZI",
    }

    assert tt.n_qubits == len(tt.root_node.child_strings)
    assert np.all(
        tt._build_symplectic_matrix()[1]
        == np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1],
                [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0],
            ],
            dtype=np.int8,
        )
    )

    for line in tt._build_symplectic_matrix()[1]:
        assert np.all(line == symplectic_unhash(symplectic_hash(line), len(line)))

def test_eigenvalues_with_openfermion(water_eigenvalues, water_integrals):
    # qham_zeros = InteractionOperator(0, tt.one_e_coeffs, np.zeros(tt.two_e_coeffs.shape))
    # ofop_zeros = jordan_wigner(qham_zeros)
    one_e_ints, two_e_ints = water_integrals
    qham = InteractionOperator(
        0, one_e_ints, 0.5*two_e_ints
    )
    # print(qham)
    ofop = jordan_wigner(qham)
    # print(f"diff {ofop-ofop_zeros}")
    diag, _ = sp.sparse.linalg.eigsh(get_sparse_operator(ofop), k=6, which="SA")

    assert np.allclose(sorted(diag), sorted(water_eigenvalues))


def test_eigenvalues_across_encodings(water_eigenvalues, water_tt, water_integrals):
    one_e_ints, two_e_ints = water_integrals

    qham2 = molecular_hamiltonian(water_tt.JKMN(), one_e_ints, 0.5*two_e_ints, 0)
    ofop2 = QubitOperator()
    for k, v in qham2.items():
        string = " ".join(
            [
                f"{char.upper()}{pos}" if char != "I" else ""
                for pos, char in enumerate(k)
            ]
        )
        ofop2 += QubitOperator(term=string, coefficient=v)
    diag2, _ = sp.sparse.linalg.eigsh(get_sparse_operator(ofop2), k=6, which="SA")

    assert np.allclose(sorted(water_eigenvalues), sorted(diag2))


def test_default_mode_op_map(water_tt):
    assert np.all(water_tt.default_mode_op_map == [*range(water_tt.n_qubits)])



"""Shared Fixtures for tests."""

import pickle
from pytest import fixture
from ferrmion.encode import TernaryTree
from pathlib import Path
from openfermion import InteractionOperator, jordan_wigner, get_sparse_operator
from scipy.sparse.linalg import eigsh

@fixture(scope="module")
def water_integrals():
    folder = Path(__file__).parent
    with open(folder.joinpath("./data/water_1e.pkl"), 'rb') as file:
        ones = pickle.load(file)

    with open(folder.joinpath("./data/water_2e.pkl"), 'rb') as file:
        twos = pickle.load(file)
    return (ones, twos)

@fixture(scope="module")
def water_tt(water_integrals) -> TernaryTree:
    return TernaryTree.from_hamiltonian_coefficients(water_integrals)

# @fixture(scope="module")
# def water_MaxNTO(water_integrals) -> MaxNTO:
#     return MaxNTO(*water_integrals)

@fixture(scope="module")
def water_eigenvalues(water_integrals) -> list[float]:
    qham = InteractionOperator(
        0, water_integrals[0], 0.5*water_integrals[1]
    )
    # print(qham)
    ofop = jordan_wigner(qham)
    # print(f"diff {ofop-ofop_zeros}")
    diag, _ = eigsh(get_sparse_operator(ofop), k=6, which="SA")
    return diag



"""Base FermionQubitEncoding class."""

import logging
from abc import ABC, abstractmethod
from itertools import product

import numpy as np
from numpy.typing import NDArray

from ferrmion.core import hartree_fock_state, symplectic_product_map
from ferrmion.utils import (
    icount_to_sign,
    pauli_to_symplectic,
    symplectic_to_pauli,
    symplectic_to_sparse,
)

logger = logging.getLogger(__name__)


class FermionQubitEncoding(ABC):
    """Fermion Encodings for the Electronic Structure Hamiltonian in symplectic form.

    Attributes:
        one_e_coeffs (NDArray): One electron coefficients.
        two_e_coeffs (NDArray): Two electron coefficients.
        modes (set[int]): A set of modes.
        n_qubits (int): The number of qubits.

    Methods:
        default_mode_op_map: Get the default mode operator map.
        _build_symplectic_matrix: Build a symplectic matrix representing terms for each operator in the Hamiltonian.
        hartree_fock_state: Find the Hartree-Fock state of a majorana string encoding.
        _symplectic_to_pauli: Convert a symplectic matrix to a Pauli string.
        _pauli_to_symplectic: Convert a Pauli string to a symplectic matrix.
        fill_template: Fill a template with Hamiltonian coefficients.
        to_symplectic_hamiltonian: Output the hamiltonian in symplectic form.
        to_qubit_hamiltonian: Create qubit representation Hamiltonian.

    NOTE: A 'Y' pauli operator is mapped to -iXY so a (0+n)**3 term is needed.

    Example:
        >>> from ferrmion.encode.base import FermionQubitEncoding
        >>> class DummyEncoding(FermionQubitEncoding):
        ...     def _build_symplectic_matrix(self):
        ...         import numpy as np
        ...         return np.zeros(1), np.zeros((1, 2))
        >>> enc = DummyEncoding(2, 2)
        >>> enc.n_modes
        2
    """

    def __init__(
        self,
        n_modes: int,
        n_qubits: int,
    ):
        """Initialise encoding.

        Args:
            n_modes (int): Number of Fermion modes to encode.
            n_qubits (int): Number of Qubits used to encode.
            vacuum_state (NDArray | None): The vacuum state of the encoding.
        """
        self.n_modes = n_modes
        self.n_qubits = n_qubits
        self.default_mode_op_map = np.array([*range(self.n_modes)], dtype=np.uint)

    def __eq__(self, other: object) -> bool:
        """Checks if two encodings are exactly equivalent."""
        if isinstance(other, FermionQubitEncoding):
            if self.n_modes != other.n_modes:
                return False

            if self.n_qubits != other.n_qubits:
                return False

            left = self._build_symplectic_matrix()
            right = other._build_symplectic_matrix()

            if not np.all(left[0] == right[0]) or not np.all(left[1] == right[1]):
                return False

            return True
        else:
            return False

    @property
    def default_mode_op_map(self):
        """Create a default mode operator map for the tree."""
        return self._default_mode_op_map

    @default_mode_op_map.setter
    def default_mode_op_map(self, permutation: list[int]):
        """Set the default mode operator map.

        Args:
            permutation (list[int]): A list containing a permutation of mode indices.
        """
        logger.debug("Setting default mode operator map.")
        error_string = ""
        if set(permutation) != {*range(self.n_modes)}:
            error_string += "Default Mode op map does not cover all modes.\n"

        if error_string != "":
            logger.error(error_string)
            logger.error(permutation)
            raise ValueError(error_string)

        self._default_mode_op_map = np.array(permutation, dtype=np.uint)

    @property
    def vacuum_state(self):
        """Return the vacuum state."""
        return self._vacuum_state

    @vacuum_state.setter
    def vacuum_state(self, state: NDArray):
        """Validate and set the vacuum state.

        Args:
            state (NDArray): The vacuum state.
        """
        logger.debug("Setting vacuum state as %s", state)
        error_string = []
        state = np.array(state, dtype=np.float64)

        if len(state) != self.n_qubits:
            error_string.append("vacuum state must be length " + str(self.n_qubits))
        if state.ndim != 1:
            error_string.append("vacuum state must be vector (dimension==1)")

        if error_string != []:
            logger.error("\n".join(error_string))
            raise ValueError("\n".join(error_string))
        else:
            self._vacuum_state = state

    @abstractmethod
    def _build_symplectic_matrix(
        self,
    ) -> tuple[NDArray[np.uint8], NDArray[bool]]:
        """Build a symplectic matrix representing terms for each operator in the Hamitonian."""
        pass

    def hartree_fock_state(
        self,
        fermionic_hf_state: NDArray[bool],
        mode_op_map: list[int] | None = None,
    ):
        """Find the Hartree-Fock state of a majorana string encoding.

        This function calls to the rust implementatin in `src/lib.rs`.
        It assumes that the vacuum state is a single state vector, though the HF state may not be
        The global phase so that the first component state has 0 phase.

        Args:
            fermionic_hf_state (NDArray[int]): An array of mode occupations.
            mode_op_map (dict[int, int]): A dictionary mapping modes to sets of majorana strings i->(j,j+1).

        Returns:
            NDArray: The Hartree-Fock ground state in computational basis.
        """
        if mode_op_map is None:
            mode_op_map = self.default_mode_op_map

        return hartree_fock_state(
            self.vacuum_state,
            fermionic_hf_state,
            mode_op_map,
            self._build_symplectic_matrix()[1],
        )

    @staticmethod
    def _symplectic_to_pauli(ipower: int, symplectic: NDArray) -> tuple[int, str]:
        """Convert a symplectic matrix to a Pauli string.

        Args:
            ipower (NDArray[np.uint]): power of i coefficient
            symplectic (NDArray): A symplectic vector.
        """
        return symplectic_to_pauli(ipower, symplectic)

    @staticmethod
    def _pauli_to_symplectic(ipower: int, pauli: str) -> tuple[int, NDArray[bool]]:
        """Convert a Pauli string to a symplectic matrix.

        Args:
            ipower (NDArray[np.uint]): power of i coefficient
            pauli (str): A Pauli-string.
        """
        return pauli_to_symplectic(ipower, pauli)

    @property
    def symplectic_product_map(self):
        """Calculate the product of symplectic terms and cache them."""
        logger.debug("Building symplectic product map")
        ipowers, symplectics = self._build_symplectic_matrix()
        return symplectic_product_map(ipowers, symplectics)

    def number_operator(
        self, mode: int
    ) -> list[tuple[str, NDArray, np.complexfloating]]:
        """Return the number operator of a mode for this encoding.

        Args:
            mode (int): The mode index to obtain a number operator for.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTee(4)
            >>> tree.number_operator(0)
        """
        return number_operator(self, mode)

    def edge_operator(
        self, edge_indices: tuple[int, int]
    ) -> list[tuple[str, NDArray, np.complexfloating]]:
        """Return the edge operator of a pair of modes for this encoding.

        Args:
            edge_indices (tuple[int, int]): The mode index to obtain a number operator for.

            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTee(4)
            >>> tree.edge_operator(0, 1)
        """
        return edge_operator(self, edge_indices)


def number_operator(
    encoding: FermionQubitEncoding, mode: int
) -> list[tuple[str, NDArray, np.complexfloating]]:
    """Return the number operator for a given encoding and mode.

    Args:
        encoding (FermionQubitEncoding): A Fermion to qubit encoding object.
        mode (int): The mode index to obtain a number operator for.

    Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTee(4)
            >>> tree.number_operator(0)
    """
    return edge_operator(encoding, (mode, mode))


def edge_operator(
    encoding: FermionQubitEncoding, edge_indices: tuple[int, int]
) -> list[tuple[str, NDArray, np.complexfloating]]:
    """Return the edge operator for a given encoding and pair of modes.

    Args:
        encoding (FermionQubitEncoding): A Fermion to qubit encoding object.
        edge_indices (tuple[int, int]): The mode index to obtain a number operator for.

    Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTee(4)
            >>> tree.edge_operator(0,1)
    """
    return double_fermionic_operator(
        encoding=encoding, mode_indices=edge_indices, signature="+-"
    )


def double_fermionic_operator(
    encoding: FermionQubitEncoding, mode_indices: tuple[int, int], signature: str
) -> list[tuple[str, NDArray, np.complexfloating]]:
    """Returns the sparse pauli form of a double fermionic operator.

    Args:
        encoding (FermionQubitEncoding): A Fermion to qubit encoding object.
        mode_indices (tuple[int, int]): The mode indices to obtain a number operator for.
        signature (str): The fermionic operator signature, one of "++", "+-", "-+", "--".

    Returns:
        list[tuple[str, NDArray, np.complexfloating]]: A list of tuples each containing a Pauli string, its qubit indices and a complex coefficient.

    Example:
        >>> from ferrmion import TernaryTree
        >>> tree = TernaryTree(4)
        >>> tree.double_fermionic_operator((0,1), "+-")
        [('ZZ', array([0, 1]), 0.25+0j), ('YX', array([0, 1]), 0.25j), ('XY', array([0, 1]), -0.25j), ('II', array([0, 1]), 0.25+0j)]
    """
    match signature:
        case "++":
            signature_iterm = [0, 3, 3, 2]
        case "+-":
            signature_iterm = [0, 1, 3, 0]
        case "-+":
            signature_iterm = [0, 3, 1, 0]
        case "--":
            signature_iterm = [0, 1, 1, 2]
        case _:
            logger.error(
                "Operator signature can only contain + or -, %s not valid", signature
            )
            raise ValueError(
                "Operator signature can only contain + or -, %s not valid", signature
            )

    logger.debug("Finding edge operator %s", mode_indices)
    if not set(mode_indices).issubset(set(range(encoding.n_modes))):
        logger.error("Edge operator indices invalid %s", mode_indices)
        raise ValueError("Edge operator indices invalid %s", mode_indices)

    icount, sym_products = encoding.symplectic_product_map
    m, n = mode_indices
    m = int(encoding.default_mode_op_map[m])
    n = int(encoding.default_mode_op_map[n])
    terms: list[tuple[int, str, NDArray]] = [
        symplectic_to_sparse(
            icount[2 * m + l, 2 * n + r], sym_products[2 * m + l, 2 * n + r]
        )
        for l, r in product([0, 1], [0, 1])
    ]
    factors = (
        0.25 * icount_to_sign(terms[0][0] + signature_iterm[0]),
        0.25 * icount_to_sign(terms[1][0] + signature_iterm[1]),
        0.25 * icount_to_sign(terms[2][0] + signature_iterm[2]),
        0.25 * icount_to_sign(terms[3][0] + signature_iterm[3]),
    )

    return [(t[1], t[2], f) for t, f in zip(terms, factors)]



"""Molecular Hamiltonian."""

import logging

from numpy.typing import NDArray

from ferrmion.core import fill_template, molecular_hamiltonian_template

from ..encode import FermionQubitEncoding

logger = logging.getLogger(__name__)


def molecular_hamiltonian(
    encoding: FermionQubitEncoding,
    one_e_coeffs: NDArray,
    two_e_coeffs: NDArray,
    constant_energy: float = 0,
    physicist_notation: bool = True,
):
    """Return an encoded electronic stucture hamiltonain with niave enumeration.

    Args:
        encoding (FermionQubitEncoding): The encoding to use.
        one_e_coeffs (NDArray): One electron hamiltonian coefficients in spinorb format.
        two_e_coeffs (NDArray): Two electron hamiltonian coefficients in spinorb format.
        constant_energy (float): Constant energy offset.
        physicist_notation (bool): Set to False for Chemist Notation.

    Example:
        >>> import numpy as np
        >>> from ferrmion.hamiltonians.molecular import molecular_hamiltonian
        >>> from ferrmion.encode import TernaryTree
        >>> tree = TernaryTree(12).JW()
        >>> one_e = np.eye((2,2))
        >>> two_e = np.eye((2,2,2,2))
        >>> molecular_hamiltonian(tree, one_e, two_e, 0.0)
    """
    ipowers, majorana_symplectic = encoding._build_symplectic_matrix()
    template = molecular_hamiltonian_template(
        ipowers, majorana_symplectic, physicist_notation
    )
    qubit_hamiltonian = fill_template(
        template=template,
        constant_energy=constant_energy,
        one_e_coeffs=one_e_coeffs,
        two_e_coeffs=two_e_coeffs,
        mode_op_map=encoding.default_mode_op_map,
    )
    return qubit_hamiltonian



"""Contains the class for ternary tree nodes."""

import logging
from typing import Optional

import rustworkx as rx

logger = logging.getLogger(__name__)


class TTNode:
    """A node in a ternary tree.

    Attributes:
        parent (TTNode): The parent node.
        label (int | str): The qubit label.
        x (TTNode): The x child node.
        y (TTNode): The y child node.
        z (TTNode): The z child node.

    Methods:
        as_dict(): Convert the node to a dictionary.
        branch_strings(): Get the branch strings for the node.
        child_strings(): Get the child strings for the node.
        add_child(which_child, root_path): Add a child node to the current node.

    Simple Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode
        >>> node = TTNode()
        >>> node.add_child('x')
        >>> node.as_dict()
    """

    def __init__(
        self,
        parent: Optional["TTNode"] = None,
        root_path: str | None = None,
        qubit_label: int | str | None = None,
    ):
        """Initialise a ternary tree node.

        Args:
            parent (TTNode | None): The parent node.
            root_path (str | None): The path from root to this node.
            qubit_label (int | str | None): The qubit label.
        """
        logger.debug(
            f"Creating TTNode with parent {parent} and qubit label {root_path}"
        )
        self.parent = parent
        self.root_path = root_path
        self.qubit_label = qubit_label
        self.x = None
        self.y = None
        self.z = None

    # def __str__(self) -> str:
    # return f"{self.as_dict()}"

    def as_dict(self) -> dict:
        """Return the node structure as a dictionary.

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x')
            >>> node.as_dict()
        """
        return as_dict(self)

    @property
    def branch_strings(self) -> set[str]:
        """Return a list of all branch strings for the node.

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x')
            >>> node.branch_strings
        """
        return branch_strings(self, prefix="")

    @property
    def child_strings(self) -> list[str]:
        """Return a list of all child strings for the node.

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x')
            >>> node.child_strings
        """
        return sorted(child_strings(self, prefix=""), key=node_sorter)

    @property
    def child_qubit_labels(self) -> dict[str, str | int | None]:
        """Return a dict of sorted child nodes and their qubit label.

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x', qubit_label=5)
            >>> node.child_qubit_labels
        """
        return child_qubit_labels(self)

    def prefix_root_path(self, prefix: str) -> None:
        """Prefix the root path of a node and all its children.

        Args:
            prefix (str): String to prefix to root paths.
        """
        self.root_path = f"{prefix}{self.root_path}"
        for child in ["x", "y", "z"]:
            child_node = getattr(self, child, None)
            if child_node is not None:
                child_node.prefix_root_path(prefix)

    def add_child(
        self,
        which_child: str,
        child_node: Optional["TTNode"] = None,
        root_path: str | None = None,
        qubit_label: int | str | None = None,
    ) -> "TTNode":
        """Add a child node to the current node.

        Args:
            which_child (str): The child node to add.
            child_node (TTNode|None): A node object to set as the child.
            root_path (str): Path from root node.
            qubit_label (int | str): The qubit label.

        Returns:
            TTNode: The added child node

        Example:
            >>> from ferrmion.encode.ternary_tree_node import TTNode
            >>> node = TTNode()
            >>> node.add_child('x')
        """
        return add_child(
            self,
            which_child=which_child,
            child_node=child_node,
            root_path=root_path,
            qubit_label=qubit_label,
        )

    def to_rustworkx(self) -> rx.PyDiGraph:
        """Create a rustworkx graph from this node and its children.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(10).BK()
            >>> rx_graph = tree.root_node.to_rustworkx()
        """
        return to_rustworkx(self)


def add_child(
    parent,
    which_child: str,
    child_node: TTNode | None = None,
    root_path: str | None = None,
    qubit_label: int | str | None = None,
) -> TTNode:
    """Add a child node to a parent node.

    Args:
        parent (TTNode): The parent node.
        which_child (str): The child node to add.
        root_path (str): Path from the root node.
        qubit_label (int | str): The qubit label.
        child_node (TTNode | None): A node to assign as child.

    Returns:
        TTNode: The added child node.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode, add_child
        >>> node = TTNode()
        >>> add_child(node, 'x')
    """
    logger.debug("Adding child %s to parent %s", which_child, parent)
    if root_path is None:
        root_path = which_child

    if (child := getattr(parent, which_child, None)) is not None:
        logger.warning(f"Already has child node {child.root_path} at {which_child}")
        pass
    elif isinstance(child_node, TTNode):
        if root_path is not None:
            child_node.prefix_root_path(parent.root_path + root_path)
        if qubit_label is not None:
            child_node.qubit_label = qubit_label
        setattr(parent, which_child, child_node)
    else:
        setattr(
            parent,
            which_child,
            TTNode(parent=parent, root_path=root_path, qubit_label=qubit_label),
        )
    return getattr(parent, which_child)


def as_dict(node: TTNode) -> dict[str, dict]:
    """Create a dictionary of children for a node.

    Args:
        node (TTNode): The node to convert to a dictionary.

    Returns:
        dict[str, dict]: A nested dictionary of children for the node.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode, as_dict
        >>> node = TTNode()
        >>> node.add_child('x')
        >>> as_dict(node)
    """
    logger.debug("Converting node to dict %s", node)
    children = {"x": node.x, "y": node.y, "z": node.z}
    for key, val in children.items():
        if val is not None:
            children[key] = as_dict(children[key])
        else:
            children[key] = {}
    return children


def child_strings(node: TTNode, prefix: str = "") -> list[str]:
    """Create a list of all child strings for a node.

    Args:
        node (TTNode): The node to convert to a list of strings.
        prefix (str): The prefix for the string.

    Returns:
        list[str]: A list of all child strings for the node.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode, child_strings
        >>> node = TTNode()
        >>> node.add_child('x')
        >>> child_strings(node)
    """
    logger.debug("Creating child strings for node %s", node)
    strings = {prefix}
    for pauli in ["x", "y", "z"]:
        child = getattr(node, pauli, None)
        if child is not None:
            strings = strings.union(child_strings(node=child, prefix=f"{prefix+pauli}"))
    logger.debug("Sorting nodes.")
    return strings


def child_qubit_labels(node: TTNode) -> dict[str, int | str | None]:
    """Return a dict of sorted child nodes and their qubit label.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode
        >>> node = TTNode()
        >>> node.add_child('x', qubit_label=5)
        >>> node.child_qubit_labels
    """
    label_dict: dict[str, int | str | None] = {}
    for child_string in node.child_strings:
        if child_string == "":
            label_dict[""] = node.qubit_label
            continue

        child: TTNode = node
        for char in child_string:
            # we are using pre-checked strings
            # so we don't need a default
            child = getattr(child, char)
        label_dict[child_string] = child.qubit_label

    return label_dict


def branch_strings(node: TTNode, prefix: str = "") -> set[str]:
    """Create a set of all branch strings for a node.

    Args:
        node (TTNode): The node to convert to a set of strings.
        prefix (str): The prefix for the string.

    Returns:
        set[str]: A set of all branch strings for the node.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import TTNode, branch_strings
        >>> node = TTNode()
        >>> node.add_child('x')
        >>> branch_strings(node)
    """
    logger.debug("Creating branch strings for node %s", node)
    strings = set()
    for pauli in ["x", "y", "z"]:
        child = getattr(node, pauli, None)
        if child is None:
            strings.add(f"{prefix+pauli}")
        else:
            strings = strings.union(
                branch_strings(node=child, prefix=f"{prefix+pauli}")
            )
    return strings


def node_sorter(label: str) -> int:
    """This is used to keep the ordring of encodings consistent.

    Args:
        label (str): The label to sort.

    Returns:
        int: Integer label to sort by.

    Example:
        >>> from ferrmion.encode.ternary_tree_node import node_sorter
        >>> node_sorter('xyz')
        123
        >>> node_sorter('xx')
        11
        >>> node_sorter('z')
        3
    """
    if label == "":
        return 0
    pauli_dict = {"x": "1", "y": "2", "z": "3"}
    return int("".join([pauli_dict[item] for item in label.lower()]))


def to_rustworkx(root: TTNode) -> rx.PyDiGraph:
    """Convert a TT node and its children to a rustworkx PyDiGraph.

    Args:
        root (TTNode): A node to be the root of the rx graph.

    Example:
        >>> from ferrmion.encode.ternary_tree import TernaryTree
        >>> tree = TernaryTree(10).BK()
        >>> rx_graph = to_rustworkx(tree.root)
    """
    graph = rx.PyDiGraph(check_cycle=True)
    child_dict = {s: i for i, s in enumerate(root.child_strings)}
    graph.add_nodes_from(child_dict)
    for string in root.child_strings:
        if len(string) == 0:
            continue
        graph.add_edge(child_dict[string[:-1]], child_dict[string], string[-1])
    return graph



"""Tests for MaxNTO encoding."""

from ferrmion.encode import MaxNTO
from ferrmion.utils import check_trivial_overlap
import numpy as np

def test_max_nto(water_integrals):
    mnto = MaxNTO(14)
    ipowers, symplectics = mnto._build_symplectic_matrix()
    valid, nto = check_trivial_overlap(symplectics)

    assert valid
    assert np.all(ipowers == np.array([0,1,0,1,2,3,2,3,0,1,0,1,2,3,2,3,0,1,0,1,2,3,2,3,0,1,0,1], dtype=np.uint8))
    assert np.all(symplectics == np.array([[False,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True, False, False, False,
        False, False, False, False, False, False, False, False, False,
        False],
       [False,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True],
       [ True, False,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True, False, False,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True],
       [ True, False,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True, False, False, False,
        False, False, False, False, False, False, False, False, False,
        False],
       [ True,  True, False,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True, False,
        False, False, False, False, False, False, False, False, False,
        False],
       [ True,  True, False,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True, False, False,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True],
       [ True,  True,  True, False,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True, False, False, False, False,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True],
       [ True,  True,  True, False,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True, False,
        False, False, False, False, False, False, False, False, False,
        False],
       [ True,  True,  True,  True, False,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True, False, False, False, False, False, False, False, False,
        False],
       [ True,  True,  True,  True, False,  True,  True,  True,  True,
         True,  True,  True,  True,  True, False, False, False, False,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True],
       [ True,  True,  True,  True,  True, False,  True,  True,  True,
         True,  True,  True,  True,  True, False, False, False, False,
        False, False,  True,  True,  True,  True,  True,  True,  True,
         True],
       [ True,  True,  True,  True,  True, False,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True, False, False, False, False, False, False, False, False,
        False],
       [ True,  True,  True,  True,  True,  True, False,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True, False, False, False, False, False, False,
        False],
       [ True,  True,  True,  True,  True,  True, False,  True,  True,
         True,  True,  True,  True,  True, False, False, False, False,
        False, False,  True,  True,  True,  True,  True,  True,  True,
         True],
       [ True,  True,  True,  True,  True,  True,  True, False,  True,
         True,  True,  True,  True,  True, False, False, False, False,
        False, False, False, False,  True,  True,  True,  True,  True,
         True],
       [ True,  True,  True,  True,  True,  True,  True, False,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True, False, False, False, False, False, False,
        False],
       [ True,  True,  True,  True,  True,  True,  True,  True, False,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True, False, False, False, False,
        False],
       [ True,  True,  True,  True,  True,  True,  True,  True, False,
         True,  True,  True,  True,  True, False, False, False, False,
        False, False, False, False,  True,  True,  True,  True,  True,
         True],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
        False,  True,  True,  True,  True, False, False, False, False,
        False, False, False, False, False, False,  True,  True,  True,
         True],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
        False,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True, False, False, False, False,
        False],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
         True, False,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True, False, False,
        False],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
         True, False,  True,  True,  True, False, False, False, False,
        False, False, False, False, False, False,  True,  True,  True,
         True],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True, False,  True,  True, False, False, False, False,
        False, False, False, False, False, False, False, False,  True,
         True],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True, False,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True, False, False,
        False],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True, False,  True,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
        False],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True, False,  True, False, False, False, False,
        False, False, False, False, False, False, False, False,  True,
         True],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True, False, False, False, False, False,
        False, False, False, False, False, False, False, False, False,
        False],
       [ True,  True,  True,  True,  True,  True,  True,  True,  True,
         True,  True,  True,  True, False,  True,  True,  True,  True,
         True,  True,  True,  True,  True,  True,  True,  True,  True,
        False]]))
    assert np.all(nto == np.array([[ 0., 13., 13.,  1.,  3., 13., 11.,  3.,  5., 11.,  9.,  5.,  7.,
         9.,  7.,  7.,  9.,  7.,  5.,  9., 11.,  5.,  3., 11., 13.,  3.,
         1., 13.],
       [13.,  0.,  1., 13., 13.,  3.,  3., 11., 11.,  5.,  5.,  9.,  9.,
         7.,  7.,  7.,  7.,  9.,  9.,  5.,  5., 11., 11.,  3.,  3., 13.,
        13.,  1.],
       [13.,  1.,  0., 13., 13.,  1.,  1., 11., 11.,  3.,  3.,  9.,  9.,
         5.,  5.,  7.,  7.,  7.,  7.,  5.,  5.,  9.,  9.,  3.,  3., 11.,
        11.,  1.],
       [ 1., 13., 13.,  0.,  1., 13., 11.,  1.,  3., 11.,  9.,  3.,  5.,
         9.,  7.,  5.,  7.,  7.,  5.,  7.,  9.,  5.,  3.,  9., 11.,  3.,
         1., 11.],
       [ 3., 13., 13.,  1.,  0., 13., 13.,  1.,  3., 13., 11.,  3.,  5.,
        11.,  9.,  5.,  7.,  9.,  7.,  7.,  9.,  7.,  5.,  9., 11.,  5.,
         3., 11.],
       [13.,  3.,  1., 13., 13.,  0.,  1., 13., 13.,  3.,  3., 11., 11.,
         5.,  5.,  9.,  9.,  7.,  7.,  7.,  7.,  9.,  9.,  5.,  5., 11.,
        11.,  3.],
       [11.,  3.,  1., 11., 13.,  1.,  0., 13., 13.,  1.,  1., 11., 11.,
         3.,  3.,  9.,  9.,  5.,  5.,  7.,  7.,  7.,  7.,  5.,  5.,  9.,
         9.,  3.],
       [ 3., 11., 11.,  1.,  1., 13., 13.,  0.,  1., 13., 11.,  1.,  3.,
        11.,  9.,  3.,  5.,  9.,  7.,  5.,  7.,  7.,  5.,  7.,  9.,  5.,
         3.,  9.],
       [ 5., 11., 11.,  3.,  3., 13., 13.,  1.,  0., 13., 13.,  1.,  3.,
        13., 11.,  3.,  5., 11.,  9.,  5.,  7.,  9.,  7.,  7.,  9.,  7.,
         5.,  9.],
       [11.,  5.,  3., 11., 13.,  3.,  1., 13., 13.,  0.,  1., 13., 13.,
         3.,  3., 11., 11.,  5.,  5.,  9.,  9.,  7.,  7.,  7.,  7.,  9.,
         9.,  5.],
       [ 9.,  5.,  3.,  9., 11.,  3.,  1., 11., 13.,  1.,  0., 13., 13.,
         1.,  1., 11., 11.,  3.,  3.,  9.,  9.,  5.,  5.,  7.,  7.,  7.,
         7.,  5.],
       [ 5.,  9.,  9.,  3.,  3., 11., 11.,  1.,  1., 13., 13.,  0.,  1.,
        13., 11.,  1.,  3., 11.,  9.,  3.,  5.,  9.,  7.,  5.,  7.,  7.,
         5.,  7.],
       [ 7.,  9.,  9.,  5.,  5., 11., 11.,  3.,  3., 13., 13.,  1.,  0.,
        13., 13.,  1.,  3., 13., 11.,  3.,  5., 11.,  9.,  5.,  7.,  9.,
         7.,  7.],
       [ 9.,  7.,  5.,  9., 11.,  5.,  3., 11., 13.,  3.,  1., 13., 13.,
         0.,  1., 13., 13.,  3.,  3., 11., 11.,  5.,  5.,  9.,  9.,  7.,
         7.,  7.],
       [ 7.,  7.,  5.,  7.,  9.,  5.,  3.,  9., 11.,  3.,  1., 11., 13.,
         1.,  0., 13., 13.,  1.,  1., 11., 11.,  3.,  3.,  9.,  9.,  5.,
         5.,  7.],
       [ 7.,  7.,  7.,  5.,  5.,  9.,  9.,  3.,  3., 11., 11.,  1.,  1.,
        13., 13.,  0.,  1., 13., 11.,  1.,  3., 11.,  9.,  3.,  5.,  9.,
         7.,  5.],
       [ 9.,  7.,  7.,  7.,  7.,  9.,  9.,  5.,  5., 11., 11.,  3.,  3.,
        13., 13.,  1.,  0., 13., 13.,  1.,  3., 13., 11.,  3.,  5., 11.,
         9.,  5.],
       [ 7.,  9.,  7.,  7.,  9.,  7.,  5.,  9., 11.,  5.,  3., 11., 13.,
         3.,  1., 13., 13.,  0.,  1., 13., 13.,  3.,  3., 11., 11.,  5.,
         5.,  9.],
       [ 5.,  9.,  7.,  5.,  7.,  7.,  5.,  7.,  9.,  5.,  3.,  9., 11.,
         3.,  1., 11., 13.,  1.,  0., 13., 13.,  1.,  1., 11., 11.,  3.,
         3.,  9.],
       [ 9.,  5.,  5.,  7.,  7.,  7.,  7.,  5.,  5.,  9.,  9.,  3.,  3.,
        11., 11.,  1.,  1., 13., 13.,  0.,  1., 13., 11.,  1.,  3., 11.,
         9.,  3.],
       [11.,  5.,  5.,  9.,  9.,  7.,  7.,  7.,  7.,  9.,  9.,  5.,  5.,
        11., 11.,  3.,  3., 13., 13.,  1.,  0., 13., 13.,  1.,  3., 13.,
        11.,  3.],
       [ 5., 11.,  9.,  5.,  7.,  9.,  7.,  7.,  9.,  7.,  5.,  9., 11.,
         5.,  3., 11., 13.,  3.,  1., 13., 13.,  0.,  1., 13., 13.,  3.,
         3., 11.],
       [ 3., 11.,  9.,  3.,  5.,  9.,  7.,  5.,  7.,  7.,  5.,  7.,  9.,
         5.,  3.,  9., 11.,  3.,  1., 11., 13.,  1.,  0., 13., 13.,  1.,
         1., 11.],
       [11.,  3.,  3.,  9.,  9.,  5.,  5.,  7.,  7.,  7.,  7.,  5.,  5.,
         9.,  9.,  3.,  3., 11., 11.,  1.,  1., 13., 13.,  0.,  1., 13.,
        11.,  1.],
       [13.,  3.,  3., 11., 11.,  5.,  5.,  9.,  9.,  7.,  7.,  7.,  7.,
         9.,  9.,  5.,  5., 11., 11.,  3.,  3., 13., 13.,  1.,  0., 13.,
        13.,  1.],
       [ 3., 13., 11.,  3.,  5., 11.,  9.,  5.,  7.,  9.,  7.,  7.,  9.,
         7.,  5.,  9., 11.,  5.,  3., 11., 13.,  3.,  1., 13., 13.,  0.,
         1., 13.],
       [ 1., 13., 11.,  1.,  3., 11.,  9.,  3.,  5.,  9.,  7.,  5.,  7.,
         7.,  5.,  7.,  9.,  5.,  3.,  9., 11.,  3.,  1., 11., 13.,  1.,
         0., 13.],
       [13.,  1.,  1., 11., 11.,  3.,  3.,  9.,  9.,  5.,  5.,  7.,  7.,
         7.,  7.,  5.,  5.,  9.,  9.,  3.,  3., 11., 11.,  1.,  1., 13.,
        13.,  0.]]))



"""Hubbard Hamiltonian."""

import numpy as np
import numpy.typing as npt

from ferrmion.core import (
    fill_template,
    hubbard_hamiltonian_template,
)

from ..encode import FermionQubitEncoding


def linear_adjacency_matrix(length: int, periodic: bool) -> npt.NDArray[bool]:
    """Creates an adjacency matrix for a linear Hubbard Hamiltonian.

    Args:
        length (int): The number of sites.
        periodic (bool): If true, periodic boundary conditions are used.

    Returns:
        np.ndarray[bool]: Adjacency matrix for lattice sites.
    """
    return square_adjacency_matrix((length, 1), periodic=periodic)


def square_adjacency_matrix(
    shape: tuple[int, int], periodic: bool
) -> npt.NDArray[bool]:
    """Creates an adjacency matrix for a 2D square lattice Hubbard Hamiltonian.

    Args:
        shape (tuple[int, int]): The number of sites.
        periodic (bool): If true, periodic boundary conditions are used.

    Returns:
        np.ndarray[bool]: Adjacency matrix for lattice sites.
    """
    # find the side length to fit nodes into square
    # we'll build a perfect square first before cutting.
    nx, ny = shape
    n_sites = nx * ny

    # initially make a chain
    adjacency_matrix = np.eye(n_sites, k=1)

    # cut chain into rows by removing connections
    for i in range(nx, n_sites, nx):
        adjacency_matrix[i - 1, i] = 0.0

    # Add connection to number below.
    adjacency_matrix += np.eye(n_sites, k=nx)

    if periodic:
        # Wrap rows
        for i in range(ny):
            adjacency_matrix[i * nx, (i + 1) * nx - 1] = 1

        # Wrap columns
        adjacency_matrix += np.eye(n_sites, k=nx * (ny - 1))

    # Hamitian conjugate
    adjacency_matrix += adjacency_matrix.T
    return np.array(adjacency_matrix, dtype=bool)


def cube_adjacency_matrix(
    shape: tuple[int, int, int], periodic: bool
) -> npt.NDArray[bool]:
    """Creates an adjacency matrix for a 3D square lattice Hubbard Hamiltonian.

    Args:
        shape (tuple[int, int, int]): The number of sites.
        periodic (bool): If true, periodic boundary conditions are used.

    Returns:
        np.ndarray[bool]: Adjacency matrix for lattice sites.
    """
    nx, ny, nz = shape
    n_sites = nx * ny * nz

    adjacency_matrix = np.zeros((n_sites, n_sites))
    # Add each of the layers of a square matrix
    for i in range(0, n_sites, nx * ny):
        adjacency_matrix[i : i + nx * ny, i : i + nx * ny] = np.triu(
            square_adjacency_matrix((nx, ny), periodic=periodic)
        )

    # Add connection in D3
    adjacency_matrix += np.eye(n_sites, k=nx * ny)

    # Wrap D3
    if periodic:
        adjacency_matrix += np.eye(n_sites, k=nx * ny * (nz - 1))

    adjacency_matrix += adjacency_matrix.T

    return np.array(adjacency_matrix, dtype=bool)


def hubbard_coefficients(
    n_modes: int,
    adjacency_matrix: npt.NDArray,
    onsite_term: float,
    hopping_term: float = 1.0,
    spinless: bool = False,
) -> tuple[np.ndarray, np.ndarray]:
    """Coefficients to fill a Hubbard Hamiltonian Template.

    Args:
        n_modes (int): Number of fermion modes in the system.
        adjacency_matrix (npt.NDArray): Adjacency matrix of lattice sites.
        onsite_term (float): Onsite interaction term.
        hopping_term (float): Kinetic term.
        spinless (bool): Set to True to use single spin Hamiltonian.

    Returns:
        tuple: one and two electron coefficients.
    """
    if not spinless:
        # We know which sites are adjacent, we need to restrict to same spin hopping.
        spin_adjacency_matrix = np.zeros(
            (2 * adjacency_matrix.shape[0], 2 * adjacency_matrix.shape[1])
        )
        spin_adjacency_matrix[::2, ::2] += adjacency_matrix
        spin_adjacency_matrix[1::2, 1::2] += adjacency_matrix
    else:
        spin_adjacency_matrix = adjacency_matrix

    one_e_coeffs = hopping_term * spin_adjacency_matrix
    one_e_coeffs = one_e_coeffs[:n_modes, :n_modes]

    two_e_coeffs = np.zeros((n_modes, n_modes, n_modes, n_modes))
    idx = np.arange(n_modes)
    two_e_coeffs[idx, idx, idx, idx] = onsite_term
    return one_e_coeffs, two_e_coeffs


def hubbard_hamiltonian(
    encoding: FermionQubitEncoding,
    adjacency_matrix: npt.NDArray,
    onsite_term: float,
    hopping_term: float = 1.0,
    spinless: bool = False,
) -> dict[str, float]:
    """Return an encoded Hubbard hamiltonain with niave enumeration.

    As the Hubbard Hamiltonian has the same signature as the Chemists' Molecular Hamiltonian:
    (+-, +-+-)
    We can use the existing functions for the molecular Hamiltonian to create a template.

    Args:
        encoding (FermionQubitEncoding): The encoding to use.
        adjacency_matrix (npt.NDArray): Adjacency matrix of lattice sites.
        onsite_term (float): Onsite two-electron term.
        hopping_term (float): Kinetic term coefficient.
        physicist_noation (bool): Set to False for Chemist Notation.
        spinless (bool): Set to True to use single spin Hamiltonian.

    Returns:
        dict[str, float]: A qubit Hamiltonian.

    Example:
        >>> import numpy as np
        >>> from ferrmion.hamiltonians.molecular import molecular_hamiltonian
        >>> from ferrmion.encode import TernaryTree
        >>> tree = TernaryTree(12).JW()
        >>> one_e = np.eye((2,2))
        >>> two_e = np.eye((2,2,2,2))
        >>> molecular_hamiltonian(tree, one_e, two_e, 0.0)
    """
    ipowers, symplectic = encoding._build_symplectic_matrix()
    template = hubbard_hamiltonian_template(ipowers, symplectic)

    n_modes = encoding.n_modes
    one_e_coeffs, two_e_coeffs = hubbard_coefficients(
        n_modes, adjacency_matrix, onsite_term, hopping_term, spinless=spinless
    )

    qubit_hamiltonian = fill_template(
        template=template,
        constant_energy=0,
        one_e_coeffs=one_e_coeffs,
        two_e_coeffs=two_e_coeffs,
        mode_op_map=encoding.default_mode_op_map,
    )
    return qubit_hamiltonian



"""Tests for base fermion to qubit encoding class"""

import numpy as np
import pytest
from ferrmion.encode import TernaryTree, MaxNTO
from ferrmion.encode.base import double_fermionic_operator

np.random.seed(1710)


@pytest.fixture
def four_mode_tt():
    return TernaryTree(n_modes=4)


@pytest.fixture
def sixteen_mode_tt():
    return TernaryTree(n_modes=16)

def test_default_vacuum_state(four_mode_tt):
    assert np.all(four_mode_tt.vacuum_state == np.array([0] * 4))

def test_valid_vacuum_state(four_mode_tt):
    with pytest.raises(ValueError) as excinfo:
        four_mode_tt.vacuum_state = [0] * 3
    assert "4" in str(excinfo.value)
    assert "length" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        four_mode_tt.vacuum_state = [0] * 5
    assert "4" in str(excinfo.value)
    assert "length" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        four_mode_tt.vacuum_state = np.array([[0], [0]])
    assert "dimension" in str(excinfo.value)


def test_hartree_fock_state(sixteen_mode_tt):
    jw = sixteen_mode_tt.JW()
    hartree_fock_state = jw.hartree_fock_state
    nq = jw.n_qubits // 2
    print(hartree_fock_state(np.array([True] * nq + [False] * nq, dtype=bool)))
    assert (
        hartree_fock_state(np.array([True] * nq + [False] * nq, dtype=bool))[0]
    ) == [1.0]
    assert np.all(
        hartree_fock_state(np.array([True] * nq + [False] * nq, dtype=bool))[1]
        == np.array([[True] * nq + [False] * nq], dtype=bool)
    )
    assert np.all(
        hartree_fock_state(
            np.array([True] * (nq + 1) + [False] * (nq - 1), dtype=bool)
        )[1]
        == np.array([[True] * (nq + 1) + [False] * (nq - 1)], dtype=bool)
    )


def test_number_operator(four_mode_tt):
    tree = four_mode_tt.JW()
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    # numpy doesn't like comparing empty arrays
    assert str(TernaryTree(n_modes=4).JW().edge_operator((0,0))) == str(TernaryTree(n_modes=4).JW().number_operator(0))
    assert str(TernaryTree(n_modes=4).JW().edge_operator((1,1))) == str(TernaryTree(n_modes=4).JW().number_operator(1))
    assert str(TernaryTree(n_modes=4).JW().edge_operator((2,2))) == str(TernaryTree(n_modes=4).JW().number_operator(2))
    assert str(TernaryTree(n_modes=4).JW().edge_operator((3,3))) == str(TernaryTree(n_modes=4).JW().number_operator(3))

    with pytest.raises(ValueError) as excinfo:
        tree.number_operator(tree.n_modes+1)
    assert "indices invalid" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        tree.number_operator(-1)
    assert "indices invalid" in str(excinfo.value)

def test_edge_operator(four_mode_tt):
    tree = four_mode_tt.JKMN()
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    left = np.array([t[2] for t in tree.edge_operator((1,0))], dtype=complex)
    right = np.array([np.conjugate(t[2]) for t in tree.edge_operator((0,1))], dtype=complex)
    assert np.all(right == left[[0,2,1,3]])
    assert np.all(left == np.array([ 0.  -0.25j,  -0.25+0.j  , 0.25+0.j  ,  0.  +0.25j]))
    assert np.all(right == np.array([ 0.  -0.25j,  0.25+0.j  , -0.25+0.j  ,  0.  +0.25j]))

    assert str(tree.edge_operator((0,3))[0]) == str(('YZX', np.array([0,1,3]), -0-0.25j))
    assert str(tree.edge_operator((0,3))[1]) == str(('YZY', np.array([0,1,3]), 0.25))
    assert str(tree.edge_operator((0,3))[2]) == str(('XZX', np.array([0,2,3]), 0.25))
    assert str(tree.edge_operator((0,3))[3]) == str(('XZY', np.array([0,2,3]), 0+0.25j))

    with pytest.raises(ValueError) as excinfo:
        tree.edge_operator((0, tree.n_modes+1))
    assert "indices invalid" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        tree.edge_operator((tree.n_modes+1, 0))
    assert "indices invalid" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        tree.number_operator((0, -1))
    assert "indices invalid" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        tree.number_operator((-1, 0))
    assert "indices invalid" in str(excinfo.value)

def test_double_fermionic_operator(four_mode_tt):
    jw_expected = [('', np.array([], dtype=np.int64), 0.25), ('Z', np.array([0]), -0.25), ('Z', np.array([0]), -0.25), ('', np.array([], dtype=np.int64), 0.25)]
    jw_num_zero =double_fermionic_operator(four_mode_tt.JW(), (0,0),"+-")
    assert jw_num_zero[0][0] == jw_num_zero[3][0] == jw_expected[0][0]
    assert type(jw_num_zero[0][1]) == type(jw_num_zero[3][1]) == type(jw_expected[0][1])
    assert len(jw_num_zero[0][1]) == len(jw_num_zero[3][1]) == len(jw_expected[0][1])
    assert jw_num_zero[0][2] == jw_num_zero[3][2] == jw_expected[0][2]
    assert np.all(jw_num_zero[1] == jw_expected[1])
    assert np.all(jw_num_zero[2] == jw_expected[2])

    bk_expected = [('', np.array([], dtype=np.int64), 0.25), ('ZZZ', np.array([0,1,3]), -0.25), ('ZZZ', np.array([0,1,3]), -0.25), ('', np.array([], dtype=np.int64), 0.25)]
    bk_num_zero =double_fermionic_operator(four_mode_tt.BK(), (0,0),"+-")
    assert np.all(l==r for l,r in zip(bk_num_zero[1], bk_expected[1]))
    assert np.all(l==r for l,r in zip(bk_num_zero[2], bk_expected[2]))

    maxnto_expected = [('', np.array([], dtype=np.int64), 0.25), ('ZZZ', np.array([1,2,3]), 0.25), ('ZZZ', np.array([1,2,3]), 0.25), ('', np.array([], dtype=np.int64), 0.25)]
    maxnot_num_zero =double_fermionic_operator(MaxNTO(4), (0,0), "+-")
    assert np.all(l==r for l,r in zip(maxnot_num_zero[1],maxnto_expected[1]))
    assert np.all(l==r for l,r in zip(maxnot_num_zero[2],maxnto_expected[2]))



import numpy as np
from ferrmion.core import symplectic_product

def test_symplectic_product():
    xyz = np.array([1, 1, 0, 0, 1, 1], dtype=bool)
    xxx = np.array([1, 1, 1, 0, 0, 0], dtype=bool)
    zzz = np.array([0, 0, 0, 1, 1, 1], dtype=bool)
    yyy = np.array([1, 1, 1, 1, 1, 1], dtype=bool)
    yzx = np.array([1, 0, 1, 1, 1, 0], dtype=bool)
    assert symplectic_product(xxx, zzz)[0] == 0
    assert np.all(symplectic_product(xxx, zzz)[1] == np.array([1, 1, 1, 1, 1, 1]))
    assert symplectic_product(zzz, xxx)[0] == 2
    assert np.all(symplectic_product(zzz, xxx)[1] == np.array([1, 1, 1, 1, 1, 1]))

    assert symplectic_product(xxx, yyy)[0] == 0
    assert np.all(symplectic_product(xxx, yyy)[1] == np.array([0, 0, 0, 1, 1, 1]))
    assert symplectic_product(yyy, xxx)[0] == 2
    assert np.all(symplectic_product(yyy, xxx)[1] == np.array([0, 0, 0, 1, 1, 1]))

    assert symplectic_product(zzz, yyy)[0] == 2
    assert np.all(symplectic_product(zzz, yyy)[1] == np.array([1, 1, 1, 0, 0, 0]))
    assert symplectic_product(yyy, zzz)[0] == 0
    assert np.all(symplectic_product(yyy, zzz)[1] == np.array([1, 1, 1, 0, 0, 0]))

    assert symplectic_product(xxx, xyz)[0] == 0
    assert np.all(symplectic_product(xxx, xyz)[1] == np.array([0, 0, 1, 0, 1, 1]))
    assert symplectic_product(xyz, xxx)[0] == 0
    assert np.all(symplectic_product(xyz, xxx)[1] == np.array([0, 0, 1, 0, 1, 1]))

    assert symplectic_product(yzx, xyz)[0] == 0
    assert np.all(symplectic_product(yzx, xyz)[1] == np.array([0, 1, 1, 1, 0, 1]))
    assert symplectic_product(xyz, yzx)[0] == 2
    assert np.all(symplectic_product(xyz, yzx)[1] == np.array([0, 1, 1, 1, 0, 1]))



"""Tests for Hamiltonan Functions."""

from ferrmion.hamiltonians import molecular_hamiltonian_template, fill_template, molecular_hamiltonian
import numpy as np
from openfermion import QubitOperator, get_sparse_operator
from scipy.sparse.linalg import eigsh
from pytest import fixture

@fixture(scope="module")
def filled_template(water_integrals, water_tt):
    symplectic_operators = water_tt.JW()._build_symplectic_matrix()
    # func_ham = molecular_hamiltonian_template(symplectic_operators[0], symplectic_operators[1])
    func_ham = molecular_hamiltonian_template(symplectic_operators[0], symplectic_operators[1], True)
    filled_template = fill_template(func_ham, 0, water_integrals[0], 0.5*water_integrals[1], water_tt.default_mode_op_map)
    return filled_template

def test_basic_molecular_hamiltonian(filled_template, water_tt, water_integrals):
    mh = molecular_hamiltonian(water_tt.JW(), water_integrals[0], water_integrals[1])
    assert filled_template.keys() == mh.keys()

def test_template(filled_template, water_eigenvalues):
    ofop3 = QubitOperator()
    for k, v in filled_template.items():
        string = " ".join(
            [
                f"{char.upper()}{pos}" if char != "I" else ""
                for pos, char in enumerate(k)
            ]
        )
        ofop3 += QubitOperator(term=string, coefficient=v)
    diag3, _ = eigsh(get_sparse_operator(ofop3), k=6, which="SA")

    assert np.allclose(sorted(diag3), sorted(water_eigenvalues))



"""Init for encodings."""

from .base import FermionQubitEncoding
from .maxnto import MaxNTO
from .ternary_tree import (
    BK,
    JKMN,
    JW,
    BravyiKitaev,
    JordanWigner,
    ParityEncoding,
    TernaryTree,
)

__all__ = [
    "FermionQubitEncoding",
    "TernaryTree",
    "MaxNTO",
    "JordanWigner",
    "JW",
    "BravyiKitaev",
    "BK",
    "ParityEncoding",
    "JKMN",
]



"""Init for enumeration optimizations."""

from .cost_functions import distance_squared, minimise_mi_distance, pauli_weighted_norm
from .evolutionary import lambda_plus_mu

__all__ = [
    "lambda_plus_mu",
    "minimise_mi_distance",
    "distance_squared",
    "pauli_weighted_norm",
]



"""Tests for Utils functions"""

import numpy as np
from ferrmion.utils import (
    icount_to_sign,
    pauli_to_symplectic,
    symplectic_hash,
    symplectic_to_pauli,
    symplectic_to_sparse,
    symplectic_unhash,
)


def test_icount_to_sign() -> None:
    assert icount_to_sign(0) == 1
    assert icount_to_sign(1) == 1j
    assert icount_to_sign(2) == -1
    assert icount_to_sign(3) == -1j


def test_symplectic_hashing() -> None:
    symplectic = np.array([0, 0, 1, 1, 0, 1, 0, 1], dtype=bool)
    print(symplectic_hash(symplectic))
    print(symplectic_unhash(symplectic_hash(symplectic), len(symplectic)))


def test_symplectic_pauli_conversion() -> None:
    symplectic = np.array([0, 0, 1, 1, 0, 1, 0, 1], dtype=bool)

    assert symplectic_to_pauli(0, symplectic) == (3, "IZXY")
    inverse_ipower, inverse_symplectic = pauli_to_symplectic(*symplectic_to_pauli(0, symplectic))
    assert inverse_ipower == 0
    assert np.all(inverse_symplectic == symplectic)


def test_symplectic_sparse_conversion() -> None:
    ipower = 1
    symplectic = np.array([0, 0, 1, 1, 0, 1, 0, 1], dtype=bool)

    assert symplectic_to_sparse(1, symplectic)[0] == 0
    assert symplectic_to_sparse(1, symplectic)[1] == "ZXY"
    assert np.array_equal(symplectic_to_sparse(1, symplectic)[2], [1,2,3])



"""Init for fermion qubit encodings.

This file is ignored by pre-commit as the pyo3 integration requires importing
rust functions before importing functions from the python module.
"""

from .core import hartree_fock_state, symplectic_product
from .encode import FermionQubitEncoding
from .encode.maxnto import MaxNTO, maxnto_symplectic_matrix
from .encode.ternary_tree import TernaryTree
from .encode.ternary_tree_node import TTNode, node_sorter
from .utils import (
    icount_to_sign,
    pauli_to_symplectic,
    setup_logs,
    symplectic_hash,
    symplectic_to_pauli,
    symplectic_unhash,
    two_operator_product,
)

__all__ = [
    "FermionQubitEncoding",
    "TernaryTree",
    "TTNode",
    "node_sorter",
    "pauli_to_symplectic",
    "symplectic_to_pauli",
    "symplectic_hash",
    "symplectic_unhash",
    "symplectic_product",
    "icount_to_sign",
    "MaxNTO",
    "maxnto_symplectic_matrix",
    "hartree_fock_state",
    "two_operator_product",
]

setup_logs()
